"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.runHolochain = exports.StateSignal = exports.StatusUpdates = exports.HOLOCHAIN_RUNNER_QUIT = exports.LAIR_KEYSTORE_QUIT = exports.ERROR_EVENT = exports.APP_PORT_EVENT = exports.STATUS_EVENT = void 0;
const childProcess = require("child_process");
const path = require("path");
const fs = require("fs");
const events_1 = require("events");
// eslint-disable-next-line @typescript-eslint/ban-ts-ignore
// @ts-ignore
const split = require("split");
const options_js_1 = require("./options.js");
const binaries_js_1 = require("./binaries.js");
const STATUS_EVENT = 'status';
exports.STATUS_EVENT = STATUS_EVENT;
const APP_PORT_EVENT = 'port';
exports.APP_PORT_EVENT = APP_PORT_EVENT;
const ERROR_EVENT = 'error';
exports.ERROR_EVENT = ERROR_EVENT;
const HOLOCHAIN_RUNNER_QUIT = 'holochain_runner_quit';
exports.HOLOCHAIN_RUNNER_QUIT = HOLOCHAIN_RUNNER_QUIT;
const LAIR_KEYSTORE_QUIT = 'lair_keystore_quit';
exports.LAIR_KEYSTORE_QUIT = LAIR_KEYSTORE_QUIT;
class StatusUpdates extends events_1.EventEmitter {
    emitStatus(status) {
        this.emit(STATUS_EVENT, status);
    }
    emitAppPort(port) {
        this.emit(APP_PORT_EVENT, port);
    }
    emitError(error) {
        this.emit(ERROR_EVENT, error);
    }
    emitHolochainRunnerQuit() {
        this.emit(HOLOCHAIN_RUNNER_QUIT);
    }
    emitLairKeystoreQuit() {
        this.emit(LAIR_KEYSTORE_QUIT);
    }
}
exports.StatusUpdates = StatusUpdates;
var StateSignal;
(function (StateSignal) {
    StateSignal[StateSignal["IsFirstRun"] = 0] = "IsFirstRun";
    StateSignal[StateSignal["IsNotFirstRun"] = 1] = "IsNotFirstRun";
    StateSignal[StateSignal["CreatingKeys"] = 2] = "CreatingKeys";
    StateSignal[StateSignal["RegisteringDna"] = 3] = "RegisteringDna";
    StateSignal[StateSignal["InstallingApp"] = 4] = "InstallingApp";
    StateSignal[StateSignal["EnablingApp"] = 5] = "EnablingApp";
    StateSignal[StateSignal["AddingAppInterface"] = 6] = "AddingAppInterface";
    StateSignal[StateSignal["IsReady"] = 7] = "IsReady";
})(StateSignal = exports.StateSignal || (exports.StateSignal = {}));
function stdoutToStateSignal(string) {
    switch (string) {
        case '0':
            return StateSignal.IsFirstRun;
        case '1':
            return StateSignal.IsNotFirstRun;
        // IsFirstRun events
        case '2':
            return StateSignal.CreatingKeys;
        case '3':
            return StateSignal.RegisteringDna;
        case '4':
            return StateSignal.InstallingApp;
        case '5':
            return StateSignal.EnablingApp;
        case '6':
            return StateSignal.AddingAppInterface;
        // Done/Ready Event
        case '7':
            return StateSignal.IsReady;
        default:
            return null;
    }
}
function checkLairInitialized(pathToLairConfig) {
    // this file should be sitting on the file system
    // if the `lair-keystore init` command has been previously run in that directory
    const configFile = path.join(pathToLairConfig, 'lair-keystore-config.yaml');
    return fs.existsSync(configFile);
}
async function runHolochain(statusEmitter, options, pathOptions) {
    const lairKeystoreBinaryPath = pathOptions
        ? pathOptions.lairKeystoreBinaryPath
        : binaries_js_1.defaultLairKeystoreBinaryPath;
    const holochainRunnerBinaryPath = pathOptions
        ? pathOptions.holochainRunnerBinaryPath
        : binaries_js_1.defaultHolochainRunnerBinaryPath;
    if (!checkLairInitialized(options.keystorePath)) {
        fs.mkdirSync(options.keystorePath, {
            recursive: true
        });
        // similar occurs in Holochain Launcher
        // https://github.com/holochain/launcher/blob/743420b717249e7c8807e04522a21288127d8d1e/crates/lair_keystore_manager/src/versions/init.rs#L13-L21
        // first we have to initialize lair-keystore
        // IF not already initialized
        // p for "piped", relates to piping in the passphrase
        const lairInitHandle = childProcess.spawn(lairKeystoreBinaryPath, ['init', '-p'], {
            cwd: options.keystorePath,
        });
        lairInitHandle.on('error', (e) => {
            console.error('there was an error during lair-keystore init -p', e);
        });
        await new Promise((resolve, reject) => {
            lairInitHandle.stdout.on('data', (chunk) => {
                if (chunk.toString().includes('lair-keystore init connection_url')) {
                    resolve();
                }
            });
            lairInitHandle.stderr.on('data', (chunk) => {
                console.error('error during lair-keystore init', chunk.toString());
                reject(new Error(chunk.toString()));
            });
            console.log('writing passphrase to `lair-keystore init -p`');
            lairInitHandle.stdin.write(options.passphrase);
            lairInitHandle.stdin.end();
        });
    }
    // p for "piped", relates to piping in the passphrase
    const lairHandle = childProcess.spawn(lairKeystoreBinaryPath, [
        'server',
        '-p',
    ], {
        cwd: options.keystorePath
    });
    // write the passphrase in
    lairHandle.stdin.write(options.passphrase);
    lairHandle.stdin.end();
    lairHandle.stdout.on('error', (error) => {
        if (lairHandle.killed)
            return;
        console.error('lair-keystore stdout err > ' + error);
        statusEmitter.emitError(error);
    });
    lairHandle.stderr.on('data', (error) => {
        if (lairHandle.killed)
            return;
        console.error('lair-keystore stderr err' + error.toString());
        statusEmitter.emitError(error);
    });
    lairHandle.on('error', (error) => {
        if (lairHandle.killed)
            return;
        console.error('lair-keystore err > ' + error.toString());
        statusEmitter.emitError(error);
    });
    lairHandle.on('close', (code) => {
        if (lairHandle.killed)
            return;
        console.log('lair-keystore closed with code: ', code);
        statusEmitter.emitLairKeystoreQuit();
    });
    // translate from ElectronHolochainOptions to HolochainRunnerOptions
    const { keystorePath, ...restOfOptions } = options;
    // get the keystoreUrl for passing to holochain-runner
    const keystoreUrl = childProcess.spawnSync(lairKeystoreBinaryPath, ['url'], {
        cwd: options.keystorePath
    }).stdout.toString();
    const holochainRunnerOptions = {
        ...restOfOptions,
        keystoreUrl
    };
    const optionsArray = options_js_1.constructOptions(holochainRunnerOptions);
    const holochainRunnerHandle = childProcess.spawn(holochainRunnerBinaryPath, optionsArray);
    // write the passphrase through stdin
    holochainRunnerHandle.stdin.write(options.passphrase);
    holochainRunnerHandle.stdin.end();
    // split divides up the stream line by line
    holochainRunnerHandle.stdout.pipe(split()).on('data', (line) => {
        console.debug('holochain > ' + line);
        // Check for state signal
        const checkIfSignal = stdoutToStateSignal(line);
        if (checkIfSignal !== null) {
            statusEmitter.emitStatus(checkIfSignal);
        }
        // Check for app port
        const appPort = parseForAppPort(line);
        if (appPort !== null) {
            statusEmitter.emitAppPort(appPort);
        }
    });
    holochainRunnerHandle.stdout.on('error', (error) => {
        if (holochainRunnerHandle.killed)
            return;
        console.error('holochain stdout err > ' + error);
        statusEmitter.emitError(error);
    });
    holochainRunnerHandle.stderr.on('data', (error) => {
        if (holochainRunnerHandle.killed)
            return;
        console.error('holochain stderr err > ' + error.toString());
        statusEmitter.emitError(new Error(error.toString()));
    });
    holochainRunnerHandle.on('error', (error) => {
        if (holochainRunnerHandle.killed)
            return;
        console.error('holochain err > ' + error.toString());
        statusEmitter.emitError(error);
    });
    holochainRunnerHandle.on('close', (code) => {
        if (holochainRunnerHandle.killed)
            return;
        console.log('holochain-runner closed with code: ', code);
        statusEmitter.emitHolochainRunnerQuit();
    });
    return {
        lairHandle,
        holochainRunnerHandle,
    };
}
exports.runHolochain = runHolochain;
function parseForAppPort(line) {
    let regex = /APP_WS_PORT: ([0-9]*)/gm;
    let match = regex.exec(line);
    // console.log({match});
    if (match === undefined || match === null || match.length === 0) {
        return null;
    }
    return match[1];
}
//# sourceMappingURL=holochain.js.map