"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getLairVersion = exports.getRunnerVersion = exports.HOLOCHAIN_RUNNER_QUIT = exports.LAIR_KEYSTORE_QUIT = exports.ERROR_EVENT = exports.APP_PORT_EVENT = exports.STATUS_EVENT = exports.StatusUpdates = exports.StateSignal = void 0;
const holochain_1 = require("./holochain");
Object.defineProperty(exports, "StateSignal", { enumerable: true, get: function () { return holochain_1.StateSignal; } });
Object.defineProperty(exports, "StatusUpdates", { enumerable: true, get: function () { return holochain_1.StatusUpdates; } });
Object.defineProperty(exports, "STATUS_EVENT", { enumerable: true, get: function () { return holochain_1.STATUS_EVENT; } });
Object.defineProperty(exports, "APP_PORT_EVENT", { enumerable: true, get: function () { return holochain_1.APP_PORT_EVENT; } });
Object.defineProperty(exports, "ERROR_EVENT", { enumerable: true, get: function () { return holochain_1.ERROR_EVENT; } });
Object.defineProperty(exports, "LAIR_KEYSTORE_QUIT", { enumerable: true, get: function () { return holochain_1.LAIR_KEYSTORE_QUIT; } });
Object.defineProperty(exports, "HOLOCHAIN_RUNNER_QUIT", { enumerable: true, get: function () { return holochain_1.HOLOCHAIN_RUNNER_QUIT; } });
const binaries_1 = require("./binaries");
const childProcess = require("child_process");
const fs = require("fs");
const kill = require("tree-kill");
// start up lair and holochain-runner processes,
// automatically shut them down on app quit,
// and emit events for status updates on their installation progress
async function initAgent(app, opts, binaryPaths) {
    // wait for the app to be ready
    await app.whenReady();
    const statusEmitter = new holochain_1.StatusUpdates();
    // execute this in a callback
    // so that we can continue and return
    // the statusEmitter to the caller
    let lairHandle;
    let holochainRunnerHandle;
    (async () => {
        let handles = await holochain_1.runHolochain(statusEmitter, opts, binaryPaths);
        lairHandle = handles.lairHandle;
        holochainRunnerHandle = handles.holochainRunnerHandle;
        app.on('will-quit', async () => {
            // SIGTERM signal is the default, and that's good
            await killHolochain(lairHandle, holochainRunnerHandle);
        });
    })();
    return {
        statusEmitter,
        shutdown: async () => {
            // SIGTERM signal is the default, and that's good
            await killHolochain(lairHandle, holochainRunnerHandle);
        },
    };
}
exports.default = initAgent;
/**
 *
 */
function sleep(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
/**
 * Kill handles and their children
 */
async function killHolochain(lairHandle, holochainRunnerHandle) {
    // Kill holochain and its children
    let canWaitForHolochain = false;
    if (holochainRunnerHandle && holochainRunnerHandle.pid) {
        canWaitForHolochain = true;
        console.debug('Killing holochain sub processes...');
        kill(holochainRunnerHandle.pid, function (err) {
            canWaitForHolochain = false;
            if (!err) {
                console.debug('killed all holochain sub processes');
            }
            else {
                console.error(err);
            }
        });
        holochainRunnerHandle.kill();
    }
    // Kill lair-keystore and its children
    let canWaitForKeystore = false;
    if (lairHandle && lairHandle.pid) {
        canWaitForKeystore = true;
        console.debug('Killing lair-keystore sub processes...');
        kill(lairHandle.pid, function (err) {
            canWaitForKeystore = false;
            if (!err) {
                console.debug('killed all lair-keystore sub processes');
            }
            else {
                console.error(err);
            }
        });
        lairHandle.kill();
    }
    // Wait for the kill commands to complete and exit anyway after a timeout
    console.debug('waiting...');
    const start_time = Date.now();
    while (canWaitForHolochain || canWaitForKeystore) {
        await sleep(10);
        if (Date.now() - start_time > 5 * 1000) {
            console.error('Killing sub-processes TIMED-OUT. Aborted.');
            break;
        }
    }
    console.debug('Killing sub-processes DONE.');
}
/**
 *
 */
function getRunnerVersion(runnerBinaryPath) {
    const holochainRunnerBinaryPath = runnerBinaryPath
        ? runnerBinaryPath
        : binaries_1.defaultHolochainRunnerBinaryPath;
    if (!fs.existsSync(holochainRunnerBinaryPath)) {
        console.error('holochain-runner binary not found at path: ' + holochainRunnerBinaryPath);
        return 'holochain-runner missing';
    }
    const holochainHandle = childProcess.spawnSync(holochainRunnerBinaryPath, [
        '--version',
    ]);
    if (holochainHandle.error) {
        console.error('Calling holochain-runner failed: ' + holochainHandle.error.message);
        return 'holochain-runner broken';
    }
    return holochainHandle.stdout.toString();
}
exports.getRunnerVersion = getRunnerVersion;
/**
 *
 */
function getLairVersion(lairBinaryPath) {
    const binaryPath = lairBinaryPath
        ? lairBinaryPath
        : binaries_1.defaultLairKeystoreBinaryPath;
    if (!fs.existsSync(binaryPath)) {
        console.error('lair-keystore binary not found at path: ' + binaryPath);
        return 'lair-keystore missing';
    }
    const handle = childProcess.spawnSync(binaryPath, ['--version']);
    if (handle.error) {
        console.error('Calling lair-keystore failed: ' + handle.error.message);
        return 'lair-keystore broken';
    }
    return handle.stdout.toString();
}
exports.getLairVersion = getLairVersion;
//# sourceMappingURL=index.js.map